import websettings


class override_websettings(object):
    def __init__(self, **kwargs):
        self.options = kwargs
        self.tmp_settings = websettings.websettings.settings.copy()

    def __enter__(self):
        for k, v in self.options.items():
            websettings.websettings.settings[k] = v

    def __exit__(self, exc_type, exc_val, exc_tb):
        websettings.websettings.settings = self.tmp_settings



from django.conf import settings
from django.utils import six
from django.utils.importlib import import_module

from websettings.backends import backend_module


class SettingStoreMetaClass(type):
    def __new__(cls, *args, **kwargs):
        super_new = super(SettingStoreMetaClass, cls).__new__
        new_class = super_new(cls, *args, **kwargs)

        mod_path = settings.WEBSETTINGS_MODULE
        mod = import_module(mod_path)
        new_class.settings = {}

        for setting in dir(mod):
            if setting == setting.upper():
                setting_value = getattr(mod, setting)
                new_class.settings[setting] = setting_value

        return new_class


class SettingStore(six.with_metaclass(SettingStoreMetaClass, object)):
    # For more testable
    backend_module = backend_module

    def __getattr__(self, item):
        if item == item.upper():
            if item not in self.settings.keys():
                raise AttributeError
            try:
                attr = self.backend_module.getsetting(item)
            except AttributeError:
                attr = self.settings[item]
            return attr

    def __setattr__(self, key, value):
        if key == key.upper():
            if key not in self.settings.keys():
                raise AttributeError
            self.backend_module.setsetting(key, value)
        else:
            return super(SettingStore, self).__setattr__(key, value)

    def __iter__(self):
        for key in self.settings.keys():
            yield key, getattr(self, key)



from websettings.models import Setting


def getsetting(item):
    try:
        s = Setting.objects.get(key=item)
    except Setting.DoesNotExist:
        raise AttributeError

    return s.value


def setsetting(key, value):
    try:
        s = Setting.objects.get(key=key)
        s.value = value
        s.save()
    except Setting.DoesNotExist:
        Setting.objects.create(key=key,
                               value=value)



from django.db import models


class Setting(models.Model):
    key = models.CharField(max_length=64)
    value = models.CharField(max_length=255)

    def __unicode__(self):
        return self.key + ': ' + self.value

    class Meta:
        db_table = 'websettings_setting'



from django.conf import settings
from django.shortcuts import HttpResponseRedirect


def admin_required(view_func):
    def wrapper(request, *args, **kwargs):
        if request.user.is_authenticated() and request.user.is_staff:
            return view_func(request, *args, **kwargs)
        else:
            return HttpResponseRedirect(settings.LOGIN_URL)

    return wrapper



from django.conf import settings
from django.utils.importlib import import_module

from websettings.backends import db as db_backend

try:
    backend_path = settings.WEBSETTINGS_BACKEND
    backend_module = import_module(backend_path)
except (AttributeError, TypeError, ImportError):
    backend_module = db_backend





from functools import partial

from django import forms
from django.utils import six
from django.utils.datastructures import SortedDict

from websettings import websettings


def get_settingstore_fields(setting_store):
    fields = [(k, forms.CharField(max_length=255,
                                  label=k,
                                  initial=partial(getattr, setting_store, k)))
              for k in setting_store.settings.keys()]
    return SortedDict(fields)


class SettingStoreFieldsMetaclass(type):
    def __new__(cls, name, bases, attrs):
        super_new = super(SettingStoreFieldsMetaclass, cls).__new__
        new_class = super_new(cls, name, bases, attrs)

        setting_store = attrs.get('setting_store', websettings)
        new_class.base_fields = get_settingstore_fields(setting_store)
        return new_class


class BaseSettingStoreForm(forms.BaseForm):
    def save(self):
        if hasattr(self, 'cleaned_data'):
            for k, v in self.cleaned_data.items():
                setattr(self.setting_store, k, v)
        else:
            # TODO: need more humanization.
            raise ValueError('Form needs cleaned_data')


class SettingStoreForm(six.with_metaclass(SettingStoreFieldsMetaclass, BaseSettingStoreForm)):
    # For more testable
    setting_store = websettings



from django.template.response import TemplateResponse
from django.shortcuts import redirect
from django.views.decorators.http import require_http_methods

from websettings import websettings
from websettings.authentication import admin_required
from websettings.forms import SettingStoreForm


@require_http_methods(["GET"])
@admin_required
def list_view(request):
    return TemplateResponse(request,
                            'websettings/list.html',
                            dict(websettings=websettings))


@require_http_methods(["GET", "POST"])
@admin_required
def edit_view(request):
    if request.method == 'GET':
        form = SettingStoreForm()
    else:
        form = SettingStoreForm(request.POST)
        if form.is_valid():
            form.save()
            return redirect('websettings.views.list_view')

    return TemplateResponse(request,
                            'websettings/edit.html',
                            dict(form=form))



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup(name='django-websettings',
      version='0.0.1',
      description="Django application to provide a web interface "
                  "to set a yet another django's settings",
      long_description=README,
      classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Web Environment",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Framework :: Django",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        ],
      author='Hiroki KIYOHARA',
      author_email='hirokiky@gmail.com',
      url='https://github.com/hirokiky/django-websettings',
      keywords='web wsgi django',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'django>=1.5',
      ],
      test_suite='runtest.main',
)



from django.test import TestCase, RequestFactory

from websettings.testing import override_websettings


__all__ = (
    'AdminRequiredTest',
    'DBBackendTest',
    'DBSettingStoreTest',
    'SettingStoreTest',
    'SettingStoreFormTest',
    'EditViewTest',
)


class DBBackendTest(TestCase):
    def _getTarget(self):
        from websettings.backends import db
        return db

    def _set_setting(self, **kwargs):
        from websettings.models import Setting
        return Setting.objects.create(**kwargs)

    def _get_setting(self, **kwargs):
        from websettings.models import Setting
        return Setting.objects.get(**kwargs)

    def test_getsetting_with_setting(self):
        self._set_setting(key='BOSS', value='ritsu')
        target = self._getTarget()
        result = target.getsetting('BOSS')
        self.assertEqual('ritsu', result)

    def test_getsetting_without_setting(self):
        target = self._getTarget()
        self.assertRaises(AttributeError, target.getsetting, 'BOSS')

    def test_setsetting_with_setting(self):
        self._set_setting(key='BOSS', value='ritsu')
        target = self._getTarget()
        target.setsetting('BOSS', 'mio')

        result = self._get_setting(key='BOSS')
        self.assertEqual(result.value, 'mio')

    def test_setting_without_setting(self):
        target = self._getTarget()
        target.setsetting('BOSS', 'ritsu')

        result = self._get_setting(key='BOSS')
        self.assertEqual(result.value, 'ritsu')


class AdminRequiredTest(TestCase):
    def _getTarget(self):
        from websettings.authentication import admin_required
        return admin_required

    def _makeOne(self):
        def dummy(request):
            return 'dummy response'
        target = self._getTarget()
        return target(dummy)

    def _get_request(self, *args, **kwargs):
        rf = RequestFactory()
        return rf.get(*args, **kwargs)

    def test_it_with_staff(self):
        from django.contrib.auth.models import User
        target = self._makeOne()
        req = self._get_request(path='/test/')
        req.user = User(username='dummy', is_staff=True)
        result = target(req)
        self.assertEqual('dummy response', result)

    def test_it_with_anonymous(self):
        from django.contrib.auth.models import AnonymousUser
        target = self._makeOne()
        req = self._get_request(path='/test/')
        req.user = AnonymousUser()
        with self.settings(LOGIN_URL='/dummy/'):
            result = target(req)
            self.assertEqual(result['Location'], '/dummy/')


class SettingStoreTest(TestCase):
    def _getTarget(self):
        from websettings.base import SettingStore
        return SettingStore

    def _makeOne(self, backend_module):
        target_class = self._getTarget()
        target = target_class()
        target.backend_module = backend_module
        return target

    def test_getattr(self):
        class Dummy(object):
            def getsetting(self, item):
                return item
        target = self._makeOne(Dummy())
        target.settings = {'DRUM': 'before'}
        self.assertEqual(target.DRUM, 'DRUM')

    def test_setattr(self):
        store = {}

        class Dummy(object):
            potal = store

            def setsetting(self, key, value):
                self.potal[key] = value
        target = self._makeOne(Dummy())
        target.settings = {'DRUM': 'before'}
        target.DRUM = 'ritsu'
        self.assertEqual(store['DRUM'], 'ritsu')


class DBSettingStoreTest(TestCase):
    def _getTarget(self):
        from websettings import websettings
        return websettings

    def test_get(self):
        target = self._getTarget()
        self.assertEqual(target.TEST_SETTING, 'before')

    def test_get_with_db(self):
        from websettings.models import Setting
        Setting.objects.create(key='TEST_SETTING', value='after')
        target = self._getTarget()
        self.assertEqual(target.TEST_SETTING, 'after')


class SettingStoreFormTest(TestCase):
    def _getTarget(self):
        from websettings.forms import SettingStoreForm
        return SettingStoreForm

    def _makeOne(self, store):
        target = self._getTarget()

        class TestSettingStoreForm(target):
            setting_store = store

        return TestSettingStoreForm()

    def test_it(self):
        from django import forms

        store = {}

        class DummySettingStore(object):
            portal = store
            settings = {'BOSS': 'ritsu'}

            def __setattr__(self, key, value):
                self.portal[key] = value

            def __getattr__(self, item):
                if item == item.upper():
                    return 'dummy'
                else:
                    return super(DummySettingStore, self).__getattribute__(item)

        dummy = DummySettingStore()
        target = self._makeOne(dummy)

        target_field = target.fields['BOSS']
        self.assertTrue(isinstance(target_field, forms.CharField))
        self.assertEqual(255, target_field.max_length)
        self.assertEqual('BOSS', target_field.label)
        self.assertEqual('dummy', target_field.initial())

        target.cleaned_data = {'BOSS': 'mio'}
        target.save()

        self.assertEqual('mio', store['BOSS'])


class EditViewTest(TestCase):
    def _getTarget(self):
        from websettings.views import edit_view
        return edit_view

    def test_get(self):
        from django.contrib.auth.models import User

        rf = RequestFactory()
        request = rf.get('/dummy')
        request.user = User(username='dummy', is_staff=True)

        with override_websettings(BOSS='ritsu'):
            from websettings.forms import SettingStoreForm
            target = self._getTarget()

            result = target(request)
            form = result.context_data['form']

            self.assertTrue(isinstance(form, SettingStoreForm))
            self.assertEqual({'TEST_SETTING', 'BOSS'},
                             set(form.fields.keys()))
            self.assertTemplateUsed(result, 'websettings/edit.html')

    def test_post(self):
        from django.contrib.auth.models import User

        rf = RequestFactory()
        request = rf.post('/dummy',
                          {'TEST_SETTING': 'after',
                           'BOSS': 'mio'})
        request.user = User(username='dummy', is_staff=True)

        with override_websettings(BOSS='ritsu'):
            from websettings import websettings
            target = self._getTarget()
            result = target(request)

            self.assertEqual(websettings.TEST_SETTING, 'after')
            self.assertEqual(websettings.BOSS, 'mio')
            self.assertEqual(302, result.status_code)
